
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.92'
version = '1.2.92'
full_version = '1.2.92'
git_revision = '5b243fed3685d658ee5b891c93899995b61d7c42'
release = True

if not release:
    version = full_version
    short_version += ".dev"
